"""An image engine implements image processing operators and an abstract image format.

Using these abstract classes allows the user to write image processing code that can run with various different image engines (eg imagej or opencv).
"""

import abc

ABC = abc.ABCMeta('ABC', (object,), {})

class Aabb(object):
    """Axis-Aligned bounding box"""

    def __init__(self, x_min, y_min, x_max, y_max):
        self.x_min = x_min
        self.y_min = y_min
        self.x_max = x_max
        self.y_max = y_max

    @property
    def width(self):
        return self.x_max - self.x_min + 1
        
    @property
    def height(self):
        return self.y_max - self.y_min + 1

class PixelType:
    U8 = 1
    U16 = 2
    F32 = 3

    PIXEL_TYPE_TO_NUM_BITS = {U8: 8,
        U16: 16,
        F32:32}

    @staticmethod
    def get_num_bits(pixel_type):
        return PixelType.PIXEL_TYPE_TO_NUM_BITS[ pixel_type ]

class Match(object):

    def __init__(self, x, y, correlation):
        self.x = x
        self.y = y
        self.correlation = correlation

    def __str__(self):
        return "(%d, %d) - %f" % (self.x, self.y, self.correlation)

class IImage(ABC):

    @abc.abstractmethod
    def width(self):
        pass

    @abc.abstractmethod
    def height(self):
        pass

    @abc.abstractmethod
    def get_subimage(self, aabb):
        """
        :param Aabb aabb: axis-aligned bounding box of the subimage
        """
        pass

    @abc.abstractmethod
    def get_pixel_type(self):
        """
        :rtype PixelType:
        """
        pass
     

class IHyperStack(ABC):

    @abc.abstractmethod
    def width(self):
        pass

    @abc.abstractmethod
    def height(self):
        pass

    @abc.abstractmethod
    def num_slices(self):
        pass

    @abc.abstractmethod
    def num_frames(self):
        pass

    @abc.abstractmethod
    def num_channels(self):
        pass


    @abc.abstractmethod
    def set_image(self, image, frame_index=0, slice_index=0, channel_index=0):
        """
        :param IImage image:
        """
        pass

    @abc.abstractmethod
    def get_image(self, frame_index=0, slice_index=0, channel_index=0):
        """
        :param IImage image:
        """
        pass


class IImageFeeder(ABC):
    """An object that generates a collection of images of the same size and same format
    """

    @abc.abstractmethod
    def __iter__(self):
        """
        for iterator
        """
        pass 

    @abc.abstractmethod
    def next(self):
        """returns the nex image in the collection
        for iterator
        """
        pass


class FileImageFeeder(IImageFeeder):

    def __init__(self):
        self.image_filepaths = []
        self.next_image_index = 0

    def __iter__(self):
        self.next_image_index = 0
        return self

    def next(self):
        if self.next_image_index < len(self.image_filepaths):
            image_filepath = self.image_filepaths[self.next_image_index]
            self.next_image_index += 1
            image = IImageEngine.get_instance().load_image(image_filepath)
            return image
        raise StopIteration

    def add_image(self, image_filepath):
        self.image_filepaths.append(image_filepath)


class StackImageFeeder(IImageFeeder):

    def __init__(self, hyperstack):
        """
        :param IHyperStack hyperstack
        """
        self.hyperstack = hyperstack

    def _init_iter(self):
        self.next_frame_index = 0
        self.next_slice_index = 0
        self.next_channel_index = 0
        self.end_is_reached = False

    def __iter__(self):
        self._init_iter()
        return self

    def next(self):
        print(self.next_channel_index, self.next_frame_index, self.next_slice_index)
        print(self.hyperstack.num_channels, self.hyperstack.num_frames, self.hyperstack.num_slices)
        if self.end_is_reached:
            raise StopIteration
        else:
            image = self.hyperstack.get_image(frame_index=self.next_frame_index, slice_index=self.next_slice_index, channel_index=self.next_channel_index)
        
        # compute next image index
        if self.next_slice_index < self.hyperstack.num_slices():
            self.next_slice_index += 1
        else:
            self.next_slice_index = 0
            if self.next_frame_index < self.hyperstack.num_frames():
                self.next_frame_index += 1
            else:
                self.next_frame_index = 0
                if self.next_channel_index < self.hyperstack.num_channels():
                    self.next_channel_index += 1
                else:
                    self.end_is_reached = True
        
        return image

    def add_image(self, image_filepath):
        self.image_filepaths.append(image_filepath)

    


class IImageEngine(ABC):
    """
    The active image processor.

    As there should only be one image engine instance at any given time, this class is implemented as a singleton
    """

    # Here will be the instance stored.
    __instance = None

    @staticmethod
    def set_instance(image_engine):
        """Change the active image_engine.

        :param IImageEngine image_engine:
        """
        IImageEngine.__instance = image_engine

    @staticmethod
    def get_instance():
        """Static access method.
        :rtype IImageEngine:
        """

        assert IImageEngine.__instance is not None
        return IImageEngine.__instance

    @abc.abstractmethod
    def create_hyperstack(self, width, height, num_channels, num_slices, num_frames, pixel_type):
        """
        :rtype IHyperStack:
        """

    @abc.abstractmethod
    def create_image(self, width, height, pixel_type):
        """
        :rtype IImage:
        """

    @abc.abstractmethod
    def load_image(self, src_image_file_path):
        """
        :param str src_image_file_path: eg './white_estimate.tiff'
        :return IImage:
        """

    @abc.abstractmethod
    def save_as_tiff(self, image, out_file_path):
        """
        :param IImage image:
        :param str out_file_path: eg './white_estimate.tiff'
        """

    @abc.abstractmethod
    def compute_max(self, image_feeder):
        """Compute for each pixel position the maximum at this position in all input images.

        :param IImageFeeder image_feeder:
        :rtype IImage:
        """

    @abc.abstractmethod
    def mean_filter(self, image, radius):
        """Each pixel becomes an average of its neighbours within the given radius

        :param IImage image:
        :param int radius: in pixels
        """

    @abc.abstractmethod
    def perform_gray_morphology(self, image, operator, structuring_element_shape, structuring_element_radius):
        """
        :param IImage image:
        :param str operator: eg 'open'
        :param str structuring_element_shape: eg 'square'
        :param int structuring_element_radius:
        """

    @abc.abstractmethod
    def replace_border(self, image, band_width):
        """Overwrites the outer band of the image by duplicating the value of the pixels that touch this outer band

        :param IImage image:
        :param int band_width: width of the outer band, in pixels
        :rtype IImage:

        adaptation for the following code from matlab telemosToolbx's estimatedwhiteFluoImageTelemos function
            outer = white_estimate;
            white_estimate = white_estimate(4:(end-3),4:(end-3));
            outer(1:3,4:(end-3))=repmat(white_estimate(1,:),3,1);  # top
            outer((end-2):end,4:(end-3))=repmat(white_estimate(end,:),3,1); # bottom
            outer(:,1:3)=repmat(outer(:,4),1,3); # left
            outer(:,(end-2):end)=repmat(outer(:,end-3),1,3); # right
            white_estimate = outer;
            clear outer;
        """
        raise NotImplementedError()
        return image

    @abc.abstractmethod
    def match_template(self, src_image, template_image):
        """
        :param IImage src_image:
        :param IImage template_image:
        :return IImage: similarity image
        """

    @abc.abstractmethod
    def find_maxima(self, src_image):
        """
        :param IImage src_image:
        :return list(Match): maxima
        """
