"""Image processing using imagej.
"""

from ..imageengine import IImage, IHyperStack, IImageEngine, PixelType, Match
from ij import IJ, ImagePlus  # pylint: disable=import-error
from ij.measure import ResultsTable
from ij.plugin import ImageCalculator  # pylint: disable=import-error
from ij.plugin.filter import MaximumFinder
from ijopencv.ij import ImagePlusMatConverter  # pylint: disable=import-error
from ijopencv.opencv import MatImagePlusConverter  # pylint: disable=import-error
import org.bytedeco.javacpp.opencv_core as opencv_core  # pylint: disable=import-error
import org.bytedeco.javacpp.opencv_imgproc as opencv_imgproc  # pylint: disable=import-error
# from org.bytedeco.javacpp.opencv_imgcodecs import imread, imwrite  # pylint: disable=import-error


IJ_PIXELTYPE_TO_PIXEL_TYPE = {
    ImagePlus.GRAY8: PixelType.U8,
    ImagePlus.GRAY16: PixelType.U16,
    ImagePlus.GRAY32: PixelType.F32
    }

class IJImage(IImage):

    def __init__(self, image_engine, ij_image=None, width=None, height=None, pixel_type=None):
        """
        :param IJImageEngine image_engine:
        :param ImagePlus ij_image:
        """
        self.image_engine = image_engine
        if ij_image is not None:
            assert ij_image.getNSlices() == 1
            assert ij_image.getNFrames() == 1
            assert ij_image.getNChannels() == 1
            self.ij_image = ij_image
        else:
            assert isinstance(width, int)
            assert isinstance(height, int)
            assert isinstance(pixel_type, int)
            stack_name = ''
            self.ij_image = IJ.createHyperStack(stack_name, width, height, 1, 1, 1, PixelType.get_num_bits(pixel_type))

    def width(self):
        return self.ij_image.width

    def height(self):
        return self.ij_image.height

    def get_pixel_type(self):
        ij_pixel_type = self.ij_image.getType()
        return IJ_PIXELTYPE_TO_PIXEL_TYPE[ij_pixel_type]

    def get_subimage(self, aabb):
        self.ij_image.saveRoi()
        self.ij_image.setRoi(aabb.x_min, aabb.y_min, aabb.width, aabb.height)
        ij_subimage = self.ij_image.crop()
        self.ij_image.restoreRoi()

        # ij_subimage = self.image_engine.create_hyperstack(width=aabb.width, height=aabb.height, num_channels=1, num_slices=1, num_frames=1, pixel_type=self.get_pixel_type())
        # ij_subimage.paste()
        return IJImage(self.image_engine, ij_subimage)


class IJHyperStack(IHyperStack):

    def __init__(self, image_engine, width, height, num_channels, num_slices, num_frames, pixel_type):
        """
        """
        self.image_engine = image_engine
        stack_name = ''
        self.hyperstack = IJ.createHyperStack(stack_name, width, height, num_channels, num_slices, num_frames, PixelType.get_num_bits(pixel_type))

    def set_image(self, image, frame_index=0, slice_index=0, channel_index=0):
        self.hyperstack.setPositionWithoutUpdate(channel_index + 1, slice_index + 1, frame_index + 1)
        self.hyperstack.setProcessor(image.ij_image.getProcessor())

    def get_image(self, frame_index=0, slice_index=0, channel_index=0):
        self.hyperstack.setPositionWithoutUpdate(channel_index + 1, slice_index + 1, frame_index + 1)
        stack_name = '' 
        image_plus = IJ.createHyperStack(stack_name, self.width(), self.height(), 1, 1, 1, PixelType.get_num_bits(self.get_pixel_type()))
        image_plus.setProcessor(self.hyperstack.getProcessor())
        ij_image = IJImage(self.image_engine, image_plus)
        return ij_image

    def width(self):
        return self.hyperstack.width

    def height(self):
        return self.hyperstack.height

    def num_slices(self):
        return self.hyperstack.getNSlices()

    def num_frames(self):
        return self.hyperstack.getNFrames()

    def num_channels(self):
        return self.hyperstack.getNChannels()

    def get_pixel_type(self):
        ij_pixel_type = self.hyperstack.getType()
        
        return IJ_PIXELTYPE_TO_PIXEL_TYPE[ij_pixel_type]



def imagej_run_image_command(image, command, options):
    """performs the given imagej command on the given image

    :param ImagePlus image:
    :param str command: imagej command (eg "Gray Morphology")
    :param str options: imagej options (eg "radius=1 type=square operator=open")

    wrapper around IJ.run (https://imagej.nih.gov/ij/developer/api/ij/IJ.html#run-ij.ImagePlus-java.lang.String-java.lang.String-) which raises an exception on error
    """
    IJ.run(image, command, options)
    error_message = IJ.getErrorMessage()
    if error_message is not None:
        raise Exception('The command "%s" with options "%s" failed because of the following error : %s' % (command, options, error_message))


def perform_gray_morphology_with_imagej(image, operator, structuring_element_shape, structuring_element_radius):
    """
    :param IJImage image:
    :param str operator: eg 'open'
    :param str structuring_element_shape: eg 'square'
    :param int structuring_element_radius:
    """
    assert operator not in ['fast open', 'fast erode'], "as of 13/09/2019, fast operators such as 'fast erode' seem broken in fiji (the resulting image is not at all similar to their slow equivalent)"
    
    processor = image.ij_image.getProcessor()
    convert_to_byte = False
    if processor.getBitDepth() != 8:
        convert_to_byte = True
        min_value = processor.getMin()
        max_value = processor.getMax()
        print("warning: downgrading image to byte as imagej's Gray Morphology processing doesn't handle 16bit images (range=[%d, %d])" % (min_value, max_value))
        do_scaling = True
        image.ij_image.setProcessor(processor.convertToByte(do_scaling))
    print("before gray morphology")
    assert structuring_element_radius < 11, "the radius of the structuring element is too big (%d); using it with Fiji's 'Gray Morphology' tool would result in very long computations." % structuring_element_radius
    imagej_run_image_command(image.ij_image, "Gray Morphology", "radius=%d type=%s operator=%s" % (structuring_element_radius, structuring_element_shape, operator))
    print("after gray morphology")


def perform_gray_morphology_with_ijopencv(image, operator, structuring_element_shape, structuring_element_radius):
    """Perform gray morphology on the given image using imagej's opencv api (this api is provided by javacpp-presets)

    :param IJImage image:
    :param str operator: eg 'open'
    :param str structuring_element_shape: eg 'square'
    :param int structuring_element_radius:
    """

    shape_name_to_opencv_id = {
        'square': opencv_imgproc.MORPH_RECT,
    }

    operator_name_to_opencv_id = {
        'open': opencv_imgproc.MORPH_OPEN,
        'close': opencv_imgproc.MORPH_CLOSE,
    }

    if structuring_element_shape not in shape_name_to_opencv_id.keys():
        raise NotImplementedError('structuring element shape %s is not currently supported (supported operators : %s)' % (structuring_element_shape, str(operator_name_to_opencv_id.keys())))

    if operator not in operator_name_to_opencv_id.keys():
        raise NotImplementedError('operator %s is not currently supported (supported operators : %s)' % (operator, str(operator_name_to_opencv_id.keys())))

    imp2mat = ImagePlusMatConverter()
    cv_src_image = imp2mat.toMat(image.ij_image.getProcessor())

    # cv_src_image = opencv_core.Mat().zeros(128, 128, opencv_core.CV_8U).asMat()
    print(cv_src_image)

    struct_el_size = structuring_element_radius * 2 + 1

    struct_el_size = opencv_core.Size(struct_el_size, struct_el_size)
    struct_el_anchor = opencv_core.Point(structuring_element_radius, structuring_element_radius)
    # print(struct_el_anchor)
    struct_element = opencv_imgproc.getStructuringElement(shape_name_to_opencv_id[structuring_element_shape], struct_el_size, struct_el_anchor)
    print(struct_element)
    # dst_image = opencv_core.Mat().zeros(128, 128, opencv_core.CV_8U).asMat()
    cv_dst_image = opencv_core.Mat(cv_src_image.size(), cv_src_image.type())

    # warning ! trying to guess the arguments from opencv's documentation is time consuming as the error messages are misleading (in the following call, javacpp complains that the 1st argument is not a org.bytedeco.javacpp.opencv_core$Mat, while it is ! The problem comes from the order of the arguments). So, instead of guessing, the found accepted signatures of morphologyEx are in javacpp-presets/opencv/src/gen/java/org/bytedeco/opencv/global/opencv_imgproc.java
    # opencv_imgproc.morphologyEx(cv_src_image, opencv_imgproc.MORPH_OPEN, struct_element, dst_image)
    # TypeError: morphologyEx(): 1st arg can't be coerced to org.bytedeco.javacpp.opencv_core$GpuMat, org.bytedeco.javacpp.opencv_core$Mat, org.bytedeco.javacpp.opencv_core$UMat
    print('before opencv_imgproc.morphologyEx')
    opencv_imgproc.morphologyEx(cv_src_image, cv_dst_image, operator_name_to_opencv_id[operator], struct_element)
    print('after opencv_imgproc.morphologyEx')

    mat2imp = MatImagePlusConverter()
    image.ij_image.setProcessor(mat2imp.toImageProcessor(cv_dst_image))


class IJImageEngine(IImageEngine):

    def missing(self):
        pass

    def create_hyperstack(self, width, height, num_channels, num_slices, num_frames, pixel_type):
        return IJHyperStack(self, width, height, num_channels, num_slices, num_frames, pixel_type)

    def create_image(self, width, height, pixel_type):
        return IJImage(self, width=width, height=height, pixel_type=pixel_type)

    def load_image(self, src_image_file_path):
        image_plus = IJ.openImage(src_image_file_path)
        return IJImage(self, image_plus)

    def save_as_tiff(self, image, out_file_path):
        IJ.saveAsTiff(image.ij_image, out_file_path)

    def divide(self, numerator_image, denominator_image):
        ic = ImageCalculator()
        result_image = ic.run("Divide create float", numerator_image.ij_image, denominator_image.ij_image)
        return IJImage(self, result_image)

    def compute_max(self, image_feeder):
        #def compute_max(self, images_file_path):
        """Computes for each pixel position the maximum at this position in all input images.

        :param IImageFeeder image_feeder:
        :rtype IJImage:
        """
        it = iter(image_feeder)

        # assert len(images_file_path) > 1
        max_image = it.next().ij_image

        #max_image = IJ.openImage(images_file_path[0])
        #print('max_image', max_image)

        # for image_file_path in images_file_path[2:-1]:
        for other_image in it:
            # other_image = IJ.openImage(image_file_path)
            print('other_image', other_image)
            ic = ImageCalculator()
            ic.run("max", max_image, other_image.ij_image)
        print('max_image', max_image)
        return IJImage(self, max_image)

    def mean_filter(self, image, radius):
        """Implement interface method."""
        IJ.run(image.ij_image, "Mean...", "radius=%d" % radius)

    def perform_gray_morphology(self, image, operator, structuring_element_shape, structuring_element_radius):
        """
        :param IJImage image:
        :param str operator: eg 'open'
        :param str structuring_element_shape: eg 'square'
        :param int structuring_element_radius:
        """
        # we use opencv version because imagej's slow version is way too slow for big values of structuring_element_radius
        perform_gray_morphology_with_ijopencv(image, operator, structuring_element_shape, structuring_element_radius)

    def replace_border(self, image, band_width):
        """Implement interface method."""
        raise NotImplementedError()
        return image

    def match_template(self, src_image, template_image):

        # import org.opencv.imgproc.Imgproc;
        #
        # /*
		#  * CV_TM_SQDIFF = 0, CV_TM_SQDIFF_NORMED = 1, CV_TM_CCORR = 2,
		#  * CV_TM_CCORR_NORMED = 3, CV_TM_CCOEFF = 4, CV_TM_CCOEFF_NORMED = 5;
		#  */
        correlation_method_id = opencv_imgproc.CV_TM_SQDIFF
		# Imgproc.matchTemplate(src, tpl, correlationImage, toCvCorrelationId(correlationMethodId));
        imp2mat = ImagePlusMatConverter()
        cv_src_image = imp2mat.toMat(src_image.ij_image.getProcessor())
        cv_template_image = imp2mat.toMat(template_image.ij_image.getProcessor())
        similarity_image_size = opencv_core.Size(src_image.width()-template_image.width()+1, src_image.width()-template_image.height()+1)
        cv_similarity_image = opencv_core.Mat(similarity_image_size, cv_src_image.type())

        # warning ! trying to guess the arguments from opencv's documentation is time consuming as the error messages are misleading (in the following call, javacpp complains that the 1st argument is not a org.bytedeco.javacpp.opencv_core$Mat, while it is ! The problem comes from the order of the arguments). So, instead of guessing, the found accepted signatures of cvMatchTemplate are in https://github.com/bytedeco/javacpp-presets/blob/master/opencv/src/gen/java/org/bytedeco/opencv/global/opencv_imgproc.java


        # print(cv_src_image)
        # print(cv_template_image)
        # print(cv_similarity_image)

        if False:
            cv_src_image = opencv_core.Mat(cv_src_image.size(), opencv_core.CV_8U)
            cv_template_image = opencv_core.Mat(cv_template_image.size(), opencv_core.CV_8U)
            cv_similarity_image = opencv_core.Mat(cv_similarity_image.size(), opencv_core.CV_8U)
            print(cv_src_image)
            print(cv_template_image)
            print(cv_similarity_image)

        opencv_imgproc.matchTemplate( cv_src_image, cv_template_image, cv_similarity_image, correlation_method_id)

        if correlation_method_id in [ opencv_imgproc.CV_TM_SQDIFF, opencv_imgproc.CV_TM_SQDIFF_NORMED ]:
            scale = -1.0
            # opencv_core.multiply(cv_similarity_image, opencv_core.Scalar.all(scale), cv_similarity_image)
            # cv_similarity_image = opencv_core.Mat(similarity_image_size, cv_src_image.type())
            # cv_similarity_image = opencv_core.Mat(cv_similarity_image.size(), opencv_core.CV_8U)
            cv_similarity_image = opencv_core.multiply(cv_similarity_image, scale).asMat()
            # print(cv_similarity_image)

        similarity_image = self.create_image(
            width=src_image.width()-template_image.width()+1,
            height=src_image.height()-template_image.height()+1,
            pixel_type=PixelType.F32)

        mat2imp = MatImagePlusConverter()
        similarity_image.ij_image.setProcessor(mat2imp.toImageProcessor(cv_similarity_image))

        return similarity_image

    def find_maxima(self, src_image):
        # to understand the meaning of threshold ansd maxima, I had to look at the source code in https://github.com/imagej/imagej1/blob/master/ij/plugin/filter/MaximumFinder.java#L636
        # 1. The algorithm first builds the list of maxima candidates. Only local maxima above this threshold are considered as valid maxima candidates
        # 2. Then each maximum candidate is processed to find its neighborhood area, starting from the highest. For ecach candidate, a propagation is performed on neighbours as long as the neighbour value is within the tolerance of the maximum candidate. If another candidate is encountered as a neighbour, then this candidate is removed as it('s been absorbed by the current maximum candidate)
        threshold = -1500.0 # only consider local maxima above this value
        tolerance = 1500 # ignore local maxima if they are in the neighborhood of another maximum. The neighborhood of a maximum is defined by the area that propagates until the the difference between the pixel value and the value of the considered local maximum exceeds tolerance. The higher the tolerance, the more local maxima are removed.... 
        # the typical value of peaks is -500 and the value between peaks is below -2500

        matches = []

        output_type = MaximumFinder.LIST
        exclude_on_edges = False
        is_euclidean_distance_map = False
        max_finder = MaximumFinder()
        max_finder.findMaxima(src_image.ij_image.getProcessor(), tolerance, threshold, output_type, exclude_on_edges, is_euclidean_distance_map)
        results_table = ResultsTable.getResultsTable()
        num_points = results_table.getCounter()
        # print("number of matches : %d" % num_points)
        for match_index in range(num_points):
            x = int(results_table.getValue('X', match_index))
            y = int(results_table.getValue('Y', match_index))
            correlation = src_image.ij_image.getProcessor().getf(x, y)
            match = Match(x, y, correlation)
            # print(match)
            matches.append(match)
        return matches